@@ -60,7 +60,7 @@ class RuleManager:
       return AggregationRule(input_pattern, output_pattern, method, frequency)
 
     except ValueError:
-      log.err("Failed to parse line: %s" % line)
+      log.err("Failed to parse rule in %s, line: %s" % (self.rules_file, line))
       raise
 
 
@@ -125,7 +125,7 @@ class AggregationRule:
 
       regex_pattern_parts.append(regex_part)
 
-    regex_pattern = '\\.'.join(regex_pattern_parts)
+    regex_pattern = '\\.'.join(regex_pattern_parts) + '$'
     self.regex = re.compile(regex_pattern)
 
   def build_template(self):
@@ -136,12 +136,16 @@ def avg(values):
   if values:
     return float( sum(values) ) / len(values)
 
+def count(values):
+  if values:
+    return len(values)
 
 AGGREGATION_METHODS = {
-  'sum' : sum,
-  'avg' : avg,
-  'min' : min,
-  'max' : max,
+  'sum'   : sum,
+  'avg'   : avg,
+  'min'   : min,
+  'max'   : max,
+  'count' : count
 }
 
 # Importable singleton
