@@ -21,8 +21,9 @@ from os.path import join, dirname, normpath, exists, isdir
 from optparse import OptionParser
 from ConfigParser import ConfigParser
 
-import whisper
-from carbon import log
+from carbon import log, state
+from carbon.database import TimeSeriesDatabase
+from carbon.routers import DatapointRouter
 from carbon.exceptions import CarbonConfigException
 
 from twisted.python import usage
@@ -32,34 +33,42 @@ defaults = dict(
   USER="",
   MAX_CACHE_SIZE=float('inf'),
   MAX_UPDATES_PER_SECOND=500,
-  MAX_UPDATES_PER_SECOND_ON_SHUTDOWN=1000,
   MAX_CREATES_PER_MINUTE=float('inf'),
+  MIN_TIMESTAMP_RESOLUTION=0,
   LINE_RECEIVER_INTERFACE='0.0.0.0',
   LINE_RECEIVER_PORT=2003,
-  LINE_RECEIVER_BACKLOG=1024,
   ENABLE_UDP_LISTENER=False,
   UDP_RECEIVER_INTERFACE='0.0.0.0',
   UDP_RECEIVER_PORT=2003,
   PICKLE_RECEIVER_INTERFACE='0.0.0.0',
   PICKLE_RECEIVER_PORT=2004,
-  PICKLE_RECEIVER_BACKLOG=1024,
+  MAX_RECEIVER_CONNECTIONS=float('inf'),
   CACHE_QUERY_INTERFACE='0.0.0.0',
   CACHE_QUERY_PORT=7002,
-  CACHE_QUERY_BACKLOG=1024,
   LOG_UPDATES=True,
+  LOG_CREATES=True,
   LOG_CACHE_HITS=True,
   LOG_CACHE_QUEUE_SORTS=True,
+  DATABASE='whisper',
   WHISPER_AUTOFLUSH=False,
   WHISPER_SPARSE_CREATE=False,
   WHISPER_FALLOCATE_CREATE=False,
   WHISPER_LOCK_WRITES=False,
+  WHISPER_FADVISE_RANDOM=False,
+  CERES_MAX_SLICE_GAP=80,
+  CERES_NODE_CACHING_BEHAVIOR='all',
+  CERES_SLICE_CACHING_BEHAVIOR='latest',
+  CERES_LOCK_WRITES=False,
   MAX_DATAPOINTS_PER_MESSAGE=500,
   MAX_AGGREGATION_INTERVALS=5,
-  FORWARD_ALL=False,
+  FORWARD_ALL=True,
   MAX_QUEUE_SIZE=1000,
-  QUEUE_LOW_WATERMARK_PCT = 0.8,
+  QUEUE_LOW_WATERMARK_PCT=0.8,
+  TIME_TO_DEFER_SENDING=0.0001,
   ENABLE_AMQP=False,
+  AMQP_METRIC_NAME_IN_BODY=False,
   AMQP_VERBOSE=False,
+  AMQP_SPEC=None,
   BIND_PATTERNS=['#'],
   ENABLE_MANHOLE=False,
   MANHOLE_INTERFACE='127.0.0.1',
@@ -68,8 +77,9 @@ defaults = dict(
   MANHOLE_PUBLIC_KEY="",
   RELAY_METHOD='rules',
   REPLICATION_FACTOR=1,
-  DIVERSE_REPLICAS=False,
+  DIVERSE_REPLICAS=True,
   DESTINATIONS=[],
+  DESTINATION_PROTOCOL="pickle",
   USE_FLOW_CONTROL=True,
   USE_INSECURE_UNPICKLER=False,
   USE_WHITELIST=False,
@@ -77,11 +87,17 @@ defaults = dict(
   CARBON_METRIC_INTERVAL=60,
   CACHE_WRITE_STRATEGY='sorted',
   WRITE_BACK_FREQUENCY=None,
-  ENABLE_LOGROTATION=True,
-  LOG_LISTENER_CONNECTIONS=True,
+  MIN_RESET_STAT_FLOW=1000,
+  MIN_RESET_RATIO=0.9,
+  MIN_RESET_INTERVAL=121,
+  USE_RATIO_RESET=False,
+  LOG_LISTENER_CONN_SUCCESS=True,
+  LOG_AGGREGATOR_MISSES=True,
   AGGREGATION_RULES='aggregation-rules.conf',
   REWRITE_RULES='rewrite-rules.conf',
   RELAY_RULES='relay-rules.conf',
+  ENABLE_LOGROTATION=True,
+  METRIC_CLIENT_IDLE_TIMEOUT=None,
 )
 
 
@@ -110,11 +126,12 @@ class OrderedConfigParser(ConfigParser):
 
     result = ConfigParser.read(self, path)
     sections = []
-    for line in open(path):
-      line = line.strip()
+    with open(path) as f:
+      for line in f:
+        line = line.strip()
 
-      if line.startswith('[') and line.endswith(']'):
-        sections.append(line[1:-1])
+        if line.startswith('[') and line.endswith(']'):
+          sections.append(line[1:-1])
 
     self._ordered_sections = sections
 
@@ -207,6 +224,14 @@ class CarbonCacheOptions(usage.Options):
         settings.update(program_settings)
         settings["program"] = program
 
+        # Normalize and expand paths
+        settings["STORAGE_DIR"] = os.path.normpath(os.path.expanduser(settings["STORAGE_DIR"]))
+        settings["LOCAL_DATA_DIR"] = os.path.normpath(os.path.expanduser(settings["LOCAL_DATA_DIR"]))
+        settings["WHITELISTS_DIR"] = os.path.normpath(os.path.expanduser(settings["WHITELISTS_DIR"]))
+        settings["PID_DIR"] = os.path.normpath(os.path.expanduser(settings["PID_DIR"]))
+        settings["LOG_DIR"] = os.path.normpath(os.path.expanduser(settings["LOG_DIR"]))
+        settings["pidfile"] = os.path.normpath(os.path.expanduser(settings["pidfile"]))
+
         # Set process uid/gid by changing the parent config, if a user was
         # provided in the configuration file.
         if settings.USER:
@@ -222,29 +247,23 @@ class CarbonCacheOptions(usage.Options):
             print "Error: missing required config %s" % storage_schemas
             sys.exit(1)
 
-        if settings.WHISPER_AUTOFLUSH:
-            log.msg("Enabling Whisper autoflush")
-            whisper.AUTOFLUSH = True
-
-        if settings.WHISPER_FALLOCATE_CREATE:
-            if whisper.CAN_FALLOCATE:
-                log.msg("Enabling Whisper fallocate support")
-            else:
-                log.err("WHISPER_FALLOCATE_CREATE is enabled but linking failed.")
-
-        if settings.WHISPER_LOCK_WRITES:
-            if whisper.CAN_LOCK:
-                log.msg("Enabling Whisper file locking")
-                whisper.LOCK = True
-            else:
-                log.err("WHISPER_LOCK_WRITES is enabled but import of fcntl module failed.")
-
-        if settings.CACHE_WRITE_STRATEGY not in ('sorted', 'max', 'naive'):
+        if settings.CACHE_WRITE_STRATEGY not in ('timesorted', 'sorted', 'max', 'naive'):
             log.err("%s is not a valid value for CACHE_WRITE_STRATEGY, defaulting to %s" %
                     (settings.CACHE_WRITE_STRATEGY, defaults['CACHE_WRITE_STRATEGY']))
         else:
-            log.msg("Using %s write strategy for cache" %
-                    settings.CACHE_WRITE_STRATEGY)
+            log.msg("Using %s write strategy for cache" % settings.CACHE_WRITE_STRATEGY)
+
+        # Database-specific settings
+        database = settings.DATABASE
+        if database not in TimeSeriesDatabase.plugins:
+            print "No database plugin implemented for '%s'" % database
+            raise SystemExit(1)
+
+        database_class = TimeSeriesDatabase.plugins[database]
+        state.database = database_class(settings)
+
+        settings.CACHE_SIZE_LOW_WATERMARK = settings.MAX_CACHE_SIZE * 0.95
+
         if not "action" in self:
             self["action"] = "start"
         self.handleAction()
@@ -254,7 +273,8 @@ class CarbonCacheOptions(usage.Options):
         # are set to log to syslog, then use that instead.
         if not self["debug"]:
             if self.parent.get("syslog", None):
-                log.logToSyslog(self.parent["prefix"])
+                prefix = "%s-%s[%d]" % (program, self["instance"], os.getpid())
+                log.logToSyslog(prefix)
             elif not self.parent["nodaemon"]:
                 logdir = settings.LOG_DIR
                 if not isdir(logdir):
@@ -301,6 +321,9 @@ class CarbonCacheOptions(usage.Options):
             try:
                 pid = int(pf.read().strip())
                 pf.close()
+            except ValueError:
+                print "Failed to parse pid from pidfile %s" % pidfile
+                raise SystemExit(1)
             except IOError:
                 print "Could not read pidfile %s" % pidfile
                 raise SystemExit(1)
@@ -323,6 +346,9 @@ class CarbonCacheOptions(usage.Options):
             try:
                 pid = int(pf.read().strip())
                 pf.close()
+            except ValueError:
+                print "Failed to parse pid from pidfile %s" % pidfile
+                raise SystemExit(1)
             except IOError:
                 print "Failed to read pid from %s" % pidfile
                 raise SystemExit(1)
@@ -341,6 +367,9 @@ class CarbonCacheOptions(usage.Options):
                 try:
                     pid = int(pf.read().strip())
                     pf.close()
+                except ValueError:
+                    print "Failed to parse pid from pidfile %s" % pidfile
+                    SystemExit(1)
                 except IOError:
                     print "Could not read pidfile %s" % pidfile
                     raise SystemExit(1)
@@ -354,6 +383,18 @@ class CarbonCacheOptions(usage.Options):
                         os.unlink(pidfile)
                     except IOError:
                         print "Could not remove pidfile %s" % pidfile
+            # Try to create the PID directory
+            else:
+                if not os.path.exists(settings["PID_DIR"]):
+                    try:
+                        os.makedirs(settings["PID_DIR"])
+                    except OSError as exc: # Python >2.5
+                        if exc.errno == errno.EEXIST and os.path.isdir(settings["PID_DIR"]):
+                           pass
+                        else:
+                           raise
+
+
 
             print "Starting %s (instance %s)" % (program, instance)
 
@@ -399,11 +440,11 @@ class CarbonRelayOptions(CarbonCacheOptions):
             self["aggregation-rules"] = join(settings["CONF_DIR"], settings['AGGREGATION_RULES'])
         settings["aggregation-rules"] = self["aggregation-rules"]
 
-        if settings["RELAY_METHOD"] not in ("rules", "consistent-hashing", "aggregated-consistent-hashing"):
-            print ("In carbon.conf, RELAY_METHOD must be either 'rules' or "
-                   "'consistent-hashing' or 'aggregated-consistent-hashing'. Invalid value: '%s'" %
-                   settings.RELAY_METHOD)
-            sys.exit(1)
+        router = settings["RELAY_METHOD"]
+        if router not in DatapointRouter.plugins:
+            print ("In carbon.conf, RELAY_METHOD must be one of %s. "
+                   "Invalid value: '%s'" % (', '.join(DatapointRouter.plugins), router))
+            raise SystemExit(1)
 
 
 def get_default_parser(usage="%prog [options] <start|stop|status>"):
@@ -530,15 +571,18 @@ def read_config(program, options, **kwargs):
                         os.environ.get("GRAPHITE_STORAGE_DIR",
                                        join(graphite_root, "storage")))
 
-    # By default, everything is written to subdirectories of the storage dir.
-    settings.setdefault(
-        "PID_DIR", settings["STORAGE_DIR"])
-    settings.setdefault(
-        "LOG_DIR", join(settings["STORAGE_DIR"], "log", program))
-    settings.setdefault(
-        "LOCAL_DATA_DIR", join(settings["STORAGE_DIR"], "whisper"))
-    settings.setdefault(
-        "WHITELISTS_DIR", join(settings["STORAGE_DIR"], "lists"))
+    def update_STORAGE_DIR_deps():
+        # By default, everything is written to subdirectories of the storage dir.
+        settings.setdefault(
+            "PID_DIR", settings["STORAGE_DIR"])
+        settings.setdefault(
+            "LOG_DIR", join(settings["STORAGE_DIR"], "log", program))
+        settings.setdefault(
+            "LOCAL_DATA_DIR", join(settings["STORAGE_DIR"], "whisper"))
+        settings.setdefault(
+            "WHITELISTS_DIR", join(settings["STORAGE_DIR"], "lists"))
+
+
 
     # Read configuration options from program-specific section.
     section = program[len("carbon-"):]
@@ -549,6 +593,7 @@ def read_config(program, options, **kwargs):
 
     settings.readFrom(config, section)
     settings.setdefault("instance", options["instance"])
+    update_STORAGE_DIR_deps()
 
     # If a specific instance of the program is specified, augment the settings
     # with the instance-specific settings and provide sane defaults for
@@ -569,4 +614,5 @@ def read_config(program, options, **kwargs):
             join(settings["PID_DIR"], '%s.pid' % program))
         settings["LOG_DIR"] = (options["logdir"] or settings["LOG_DIR"])
 
+    update_STORAGE_DIR_deps()
     return settings
